@@ -35,7 +35,6 @@ import (
 	"github.com/containerd/containerd/services/server"
 	srvconfig "github.com/containerd/containerd/services/server/config"
 	"github.com/containerd/containerd/sys"
-	"github.com/containerd/containerd/tracing"
 	"github.com/containerd/containerd/version"
 	"github.com/sirupsen/logrus"
 	"github.com/urfave/cli"
@@ -69,7 +68,7 @@ func App() *cli.App {
 	app.Usage = usage
 	app.Description = `
 containerd is a high performance container runtime whose daemon can be started
-by using this command. If none of the *config*, *publish*, or *help* commands
+by using this command. If none of the *config*, *publish*, *oci-hook*, or *help* commands
 are specified, the default action of the **containerd** command is to start the
 containerd daemon in the foreground.
 
@@ -81,16 +80,16 @@ can be used and modified as necessary as a custom configuration.`
 	app.Flags = []cli.Flag{
 		cli.StringFlag{
 			Name:  "config,c",
-			Usage: "path to the configuration file",
+			Usage: "Path to the configuration file",
 			Value: filepath.Join(defaults.DefaultConfigDir, "config.toml"),
 		},
 		cli.StringFlag{
 			Name:  "log-level,l",
-			Usage: "set the logging level [trace, debug, info, warn, error, fatal, panic]",
+			Usage: "Set the logging level [trace, debug, info, warn, error, fatal, panic]",
 		},
 		cli.StringFlag{
 			Name:  "address,a",
-			Usage: "address for containerd's GRPC server",
+			Usage: "Address for containerd's GRPC server",
 		},
 		cli.StringFlag{
 			Name:  "root",
@@ -133,6 +132,16 @@ can be used and modified as necessary as a custom configuration.`
 			return err
 		}
 
+		if config.GRPC.Address == "" {
+			return fmt.Errorf("grpc address cannot be empty: %w", errdefs.ErrInvalidArgument)
+		}
+		if config.TTRPC.Address == "" {
+			// If TTRPC was not explicitly configured, use defaults based on GRPC.
+			config.TTRPC.Address = config.GRPC.Address + ".ttrpc"
+			config.TTRPC.UID = config.GRPC.UID
+			config.TTRPC.GID = config.GRPC.GID
+		}
+
 		// Make sure top-level directories are created early.
 		if err := server.CreateTopLevelDirectories(config); err != nil {
 			return err
@@ -165,16 +174,7 @@ can be used and modified as necessary as a custom configuration.`
 			log.G(ctx).WithError(w).Warn("cleanup temp mount")
 		}
 
-		if config.GRPC.Address == "" {
-			return fmt.Errorf("grpc address cannot be empty: %w", errdefs.ErrInvalidArgument)
-		}
-		if config.TTRPC.Address == "" {
-			// If TTRPC was not explicitly configured, use defaults based on GRPC.
-			config.TTRPC.Address = fmt.Sprintf("%s.ttrpc", config.GRPC.Address)
-			config.TTRPC.UID = config.GRPC.UID
-			config.TTRPC.GID = config.GRPC.GID
-		}
-		log.G(ctx).WithFields(logrus.Fields{
+		log.G(ctx).WithFields(log.Fields{
 			"version":  version.Version,
 			"revision": version.Revision,
 		}).Info("starting containerd")
@@ -186,8 +186,8 @@ can be used and modified as necessary as a custom configuration.`
 
 		// run server initialization in a goroutine so we don't end up blocking important things like SIGTERM handling
 		// while the server is initializing.
-		// As an example opening the bolt database will block forever if another containerd is already running and containerd
-		// will have to be be `kill -9`'ed to recover.
+		// As an example, opening the bolt database blocks forever if a containerd instance
+		// is already running, which must then be forcibly terminated (SIGKILL) to recover.
 		chsrv := make(chan srvResp)
 		go func() {
 			defer close(chsrv)
@@ -302,7 +302,6 @@ func applyFlags(context *cli.Context, config *srvconfig.Config) error {
 	if err := setLogFormat(config); err != nil {
 		return err
 	}
-	setLogHooks()
 
 	for _, v := range []struct {
 		name string
@@ -323,6 +322,13 @@ func applyFlags(context *cli.Context, config *srvconfig.Config) error {
 	} {
 		if s := context.GlobalString(v.name); s != "" {
 			*v.d = s
+			if v.name == "root" || v.name == "state" {
+				absPath, err := filepath.Abs(s)
+				if err != nil {
+					return err
+				}
+				*v.d = absPath
+			}
 		}
 	}
 
@@ -369,10 +375,6 @@ func setLogFormat(config *srvconfig.Config) error {
 	return nil
 }
 
-func setLogHooks() {
-	logrus.StandardLogger().AddHook(tracing.NewLogrusHook())
-}
-
 func dumpStacks(writeToFile bool) {
 	var (
 		buf       []byte
