@@ -21,9 +21,9 @@ import (
 	"crypto/tls"
 	"crypto/x509"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"net/http"
 	"net/http/httptest"
 	"strconv"
@@ -32,10 +32,10 @@ import (
 	"time"
 
 	"github.com/containerd/containerd/remotes"
+	"github.com/containerd/containerd/remotes/docker/auth"
 	digest "github.com/opencontainers/go-digest"
 	specs "github.com/opencontainers/image-spec/specs-go"
 	ocispec "github.com/opencontainers/image-spec/specs-go/v1"
-	"github.com/pkg/errors"
 )
 
 func TestHTTPResolver(t *testing.T) {
@@ -68,11 +68,15 @@ func TestBasicResolver(t *testing.T) {
 		})
 
 		base, options, close := tlsServer(wrapped)
+		authorizer := NewDockerAuthorizer(
+			WithAuthClient(options.Client),
+			WithAuthCreds(func(host string) (string, string, error) {
+				return "user1", "password1", nil
+			}),
+		)
 		options.Hosts = ConfigureDefaultRegistries(
 			WithClient(options.Client),
-			WithAuthorizer(NewAuthorizer(options.Client, func(string) (string, string, error) {
-				return "user1", "password1", nil
-			})),
+			WithAuthorizer(authorizer),
 		)
 		return base, options, close
 	}
@@ -138,6 +142,31 @@ func TestRefreshTokenResolver(t *testing.T) {
 	runBasicTest(t, "testname", withTokenServer(th, creds))
 }
 
+func TestFetchRefreshToken(t *testing.T) {
+	f := func(t *testing.T, disablePOST bool) {
+		name := "testname"
+		if disablePOST {
+			name += "-disable-post"
+		}
+		var fetchedRefreshToken string
+		onFetchRefreshToken := func(ctx context.Context, refreshToken string, req *http.Request) {
+			fetchedRefreshToken = refreshToken
+		}
+		srv := newRefreshTokenServer(t, name, disablePOST, onFetchRefreshToken)
+		runBasicTest(t, name, srv.BasicTestFunc())
+		if fetchedRefreshToken != srv.RefreshToken {
+			t.Errorf("unexpected refresh token: got %q", fetchedRefreshToken)
+		}
+	}
+
+	t.Run("POST", func(t *testing.T) {
+		f(t, false)
+	})
+	t.Run("GET", func(t *testing.T) {
+		f(t, true)
+	})
+}
+
 func TestPostBasicAuthTokenResolver(t *testing.T) {
 	th := http.HandlerFunc(func(rw http.ResponseWriter, r *http.Request) {
 		if r.Method != http.MethodPost {
@@ -563,23 +592,45 @@ func testFetch(ctx context.Context, f remotes.Fetcher, desc ocispec.Descriptor)
 	dgstr := desc.Digest.Algorithm().Digester()
 	io.Copy(dgstr.Hash(), r)
 	if dgstr.Digest() != desc.Digest {
-		return errors.Errorf("content mismatch: %s != %s", dgstr.Digest(), desc.Digest)
+		return fmt.Errorf("content mismatch: %s != %s", dgstr.Digest(), desc.Digest)
 	}
 
+	fByDigest, ok := f.(remotes.FetcherByDigest)
+	if !ok {
+		return fmt.Errorf("fetcher %T does not implement FetcherByDigest", f)
+	}
+	r2, desc2, err := fByDigest.FetchByDigest(ctx, desc.Digest)
+	if err != nil {
+		return fmt.Errorf("FetcherByDigest: faild to fetch %v: %w", desc.Digest, err)
+	}
+	if desc2.Size != desc.Size {
+		r2b, err := io.ReadAll(r2)
+		if err != nil {
+			return fmt.Errorf("FetcherByDigest: size mismatch: %d != %d (content: %v)", desc2.Size, desc.Size, err)
+		}
+		return fmt.Errorf("FetcherByDigest: size mismatch: %d != %d (content: %q)", desc2.Size, desc.Size, string(r2b))
+	}
+	dgstr2 := desc.Digest.Algorithm().Digester()
+	if _, err = io.Copy(dgstr2.Hash(), r2); err != nil {
+		return fmt.Errorf("FetcherByDigest: faild to copy: %w", err)
+	}
+	if dgstr2.Digest() != desc.Digest {
+		return fmt.Errorf("FetcherByDigest: content mismatch: %s != %s", dgstr2.Digest(), desc.Digest)
+	}
 	return nil
 }
 
 func testocimanifest(ctx context.Context, f remotes.Fetcher, desc ocispec.Descriptor) ([]ocispec.Descriptor, error) {
 	r, err := f.Fetch(ctx, desc)
 	if err != nil {
-		return nil, errors.Wrapf(err, "failed to fetch %s", desc.Digest)
+		return nil, fmt.Errorf("failed to fetch %s: %w", desc.Digest, err)
 	}
-	p, err := ioutil.ReadAll(r)
+	p, err := io.ReadAll(r)
 	if err != nil {
 		return nil, err
 	}
 	if dgst := desc.Digest.Algorithm().FromBytes(p); dgst != desc.Digest {
-		return nil, errors.Errorf("digest mismatch: %s != %s", dgst, desc.Digest)
+		return nil, fmt.Errorf("digest mismatch: %s != %s", dgst, desc.Digest)
 	}
 
 	var manifest ocispec.Manifest
@@ -659,3 +710,131 @@ func (m testManifest) RegisterHandler(r *http.ServeMux, name string) {
 		r.Handle(fmt.Sprintf("/v2/%s/blobs/%s", name, c.Digest()), c)
 	}
 }
+
+func newRefreshTokenServer(t testing.TB, name string, disablePOST bool, onFetchRefreshToken OnFetchRefreshToken) *refreshTokenServer {
+	return &refreshTokenServer{
+		T:                   t,
+		Name:                name,
+		DisablePOST:         disablePOST,
+		OnFetchRefreshToken: onFetchRefreshToken,
+		AccessToken:         "testAccessToken-" + name,
+		RefreshToken:        "testRefreshToken-" + name,
+		Username:            "testUser-" + name,
+		Password:            "testPassword-" + name,
+	}
+}
+
+type refreshTokenServer struct {
+	T                   testing.TB
+	Name                string
+	DisablePOST         bool
+	OnFetchRefreshToken OnFetchRefreshToken
+	AccessToken         string
+	RefreshToken        string
+	Username            string
+	Password            string
+}
+
+func (srv *refreshTokenServer) isValidAuthorizationHeader(s string) bool {
+	fields := strings.Fields(s)
+	return len(fields) == 2 && strings.ToLower(fields[0]) == "bearer" && (fields[1] == srv.RefreshToken || fields[1] == srv.AccessToken)
+}
+
+func (srv *refreshTokenServer) BasicTestFunc() func(h http.Handler) (string, ResolverOptions, func()) {
+	t := srv.T
+	return func(h http.Handler) (string, ResolverOptions, func()) {
+		wrapped := http.HandlerFunc(func(rw http.ResponseWriter, r *http.Request) {
+			if r.URL.Path != "/token" {
+				if !srv.isValidAuthorizationHeader(r.Header.Get("Authorization")) {
+					realm := fmt.Sprintf("https://%s/token", r.Host)
+					wwwAuthenticateHeader := fmt.Sprintf("Bearer realm=%q,service=registry,scope=\"repository:%s:pull\"", realm, srv.Name)
+					rw.Header().Set("WWW-Authenticate", wwwAuthenticateHeader)
+					rw.WriteHeader(http.StatusUnauthorized)
+					return
+				}
+				h.ServeHTTP(rw, r)
+				return
+			}
+			switch r.Method {
+			case http.MethodGet: // https://docs.docker.com/registry/spec/auth/token/#requesting-a-token
+				u, p, ok := r.BasicAuth()
+				if !ok || u != srv.Username || p != srv.Password {
+					rw.WriteHeader(http.StatusForbidden)
+					return
+				}
+				var resp auth.FetchTokenResponse
+				resp.Token = srv.AccessToken
+				resp.AccessToken = srv.AccessToken // alias of Token
+				query := r.URL.Query()
+				switch query.Get("offline_token") {
+				case "true":
+					resp.RefreshToken = srv.RefreshToken
+				case "false", "":
+				default:
+					rw.WriteHeader(http.StatusBadRequest)
+					return
+				}
+				b, err := json.Marshal(resp)
+				if err != nil {
+					rw.WriteHeader(http.StatusInternalServerError)
+					return
+				}
+				rw.WriteHeader(http.StatusOK)
+				rw.Header().Set("Content-Type", "application/json")
+				t.Logf("GET mode: returning JSON %q, for query %+v", string(b), query)
+				rw.Write(b)
+			case http.MethodPost: // https://docs.docker.com/registry/spec/auth/oauth/#getting-a-token
+				if srv.DisablePOST {
+					rw.WriteHeader(http.StatusMethodNotAllowed)
+					return
+				}
+				r.ParseForm()
+				pf := r.PostForm
+				if pf.Get("grant_type") != "password" {
+					rw.WriteHeader(http.StatusBadRequest)
+					return
+				}
+				if pf.Get("username") != srv.Username || pf.Get("password") != srv.Password {
+					rw.WriteHeader(http.StatusForbidden)
+					return
+				}
+				var resp auth.OAuthTokenResponse
+				resp.AccessToken = srv.AccessToken
+				switch pf.Get("access_type") {
+				case "offline":
+					resp.RefreshToken = srv.RefreshToken
+				case "online", "":
+				default:
+					rw.WriteHeader(http.StatusBadRequest)
+					return
+				}
+				b, err := json.Marshal(resp)
+				if err != nil {
+					rw.WriteHeader(http.StatusInternalServerError)
+					return
+				}
+				rw.WriteHeader(http.StatusOK)
+				rw.Header().Set("Content-Type", "application/json")
+				t.Logf("POST mode: returning JSON %q, for form %+v", string(b), pf)
+				rw.Write(b)
+			default:
+				rw.WriteHeader(http.StatusMethodNotAllowed)
+				return
+			}
+		})
+
+		base, options, close := tlsServer(wrapped)
+		authorizer := NewDockerAuthorizer(
+			WithAuthClient(options.Client),
+			WithAuthCreds(func(string) (string, string, error) {
+				return srv.Username, srv.Password, nil
+			}),
+			WithFetchRefreshToken(srv.OnFetchRefreshToken),
+		)
+		options.Hosts = ConfigureDefaultRegistries(
+			WithClient(options.Client),
+			WithAuthorizer(authorizer),
+		)
+		return base, options, close
+	}
+}
